###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

__all__ = [
    "LogFileMatch",
    "LogFileMatches",
    "AppLogFailureReason",
    "Job",
    "PreAnalysisResult",
    "LFNDebugInfo",
    "AggregatedLFNDebugInfo",
]

import ast
import re
from enum import auto, StrEnum
from typing import Self
from functools import partial, cached_property
from collections import defaultdict

from pydantic import BaseModel, ConfigDict, field_validator

from .classes import JobArtifacts, JobArtifactConsistencyError
from .base_reasons import (
    LogFileMatches,
    LogFileMatch,
    AppLogFailureReason,
    WorkerNodeIssue,
)

# We need to import the reasons submodule so the subclasses of AppLogFailureReason
# are defined. It's not pretty but it gives us a fairly simple way of being able
# to define new problems that are automatically integrated into the analyzer.
from . import reasons  # noqa

RE_FIND_JOB_ID = re.compile(rb"JobID: (\d{6,})\n")


class LbCondaJobsUnsupported(NotImplementedError):
    pass


class PreAnalysisResult(StrEnum):
    NO_LOG_FILE = auto()
    NO_XML_SUMMARY = auto()
    UNKNOWN_ERROR_ON_JOB_SUCCESS = auto()
    NO_EVENTS_SELECTED = auto()
    PROBABLY_OK = auto()
    NOT_IN_SUMMARY_XML = auto()
    NEEDS_GROUP_SIZE_REDUCTION = auto()
    NEEDS_LOG_ANALYSIS = auto()
    MAYBE_POST_APPLICATION_FAILURE = auto()
    NO_LOGSE_JOB_PARAMETER = auto()
    STALLED = auto()
    FAILED_INPUT_DATA_RESOLUTION = auto()
    INPUT_DATA_ALREADY_PROCESSED = auto()


class PostAnalysisResult(StrEnum):
    NEEDS_DOWNLOAD_INPUT_DATA = auto()
    KNOWN_CORRUPTION = auto()
    HIGH_MEMORY_USAGE = auto()
    EXTREMELY_HIGH_MEMORY_USAGE = auto()
    NEEDS_GROUP_SIZE_REDUCTION = auto()
    TOO_MANY_ATTEMPTS = auto()
    TOO_MUCH_MEMORY_FOR_CGROUPS = auto()


# TODO Make an dynamic enum for type


class ReasonInfo(BaseModel):
    model_config = ConfigDict(extra="forbid")

    type: str
    match: str
    explanation: str | None


class LFNDebugInfo(BaseModel):
    model_config = ConfigDict(extra="forbid")

    lfn: str
    job_id: int
    transform_id: int
    site: str | None = None
    log_url: str | None = None
    pre_analysis: PreAnalysisResult
    peak_memory: int | None = None
    maybe_reason: ReasonInfo | None = None
    post_analysis: PostAnalysisResult | None = None
    n_input_files: int | None = None

    @field_validator("maybe_reason", mode="before")
    @classmethod
    def maybe_reason_validator(cls, v):
        if isinstance(v, AppLogFailureReason):
            return ReasonInfo(
                type=v.__class__.__name__,
                match=v.match_string,
                explanation=v.explanation,
            )
        return v


class AggregatedLFNDebugInfo(LFNDebugInfo):
    occurrences: int = 1
    attempts: int = 1
    sites: dict[str, int] = {}
    log_urls: list[str] = []


class Job:
    @classmethod
    async def from_logse_lfn(cls, logse_lfn: str) -> Self:
        raise NotImplementedError()

    def __init__(
        self,
        transform_id: int,
        task_id: int,
        files: dict[str, bytes],
        *,
        job_id: int | None = None,
        url: str | None = None,
    ) -> None:
        """Describes the logs of a job which has been processed by the LHCbDIRAC Production System.

        Args:
            transform_id: The transform ID of the job.
            task_id: The task ID of the job with the transformation system.
            files: The list of LFNs of the files which have been processed by the job.
            job_id: The job ID of the job. If not given tries to guess based on the files.
        """
        self.job_id = job_id
        self.transform_id = transform_id
        self.task_id = task_id
        self.url = url
        self.files = JobArtifacts.from_files(files, transform_id=transform_id, task_id=task_id)
        self._guess_or_check_job_id()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"job_id={self.job_id}, "
            f"transform_id={self.transform_id}, "
            f"task_id={self.task_id}, "
            f"url={self.url})"
        )

    def __hash__(self) -> int:
        return hash((self.transform_id, self.task_id, self.job_id))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.transform_id == other.transform_id and self.task_id == other.task_id and self.job_id == other.job_id

    def _guess_or_check_job_id(self):
        if match := RE_FIND_JOB_ID.search(self.files.dirac_log):
            guessed_job_id = int(match.group(1))
            if self.job_id is None:
                self.job_id = guessed_job_id
            elif self.job_id != guessed_job_id:
                raise JobArtifactConsistencyError(f"Job ID mismatch: {self.job_id} != {guessed_job_id}")

    @property
    def site(self):
        site_parameter = self.files.workflow.find(".//Parameter[@name='Site']/value")
        if site_parameter is None:
            print(f"ERROR: Failed to find site parameter for {self.job_id}")
        else:
            possible_sites = site_parameter.text
            found_sites = set()
            for site in possible_sites.split(";"):
                if f"is running at site {site}\n".encode() in self.files.dirac_log:
                    found_sites.add(site)
            if len(found_sites) == 1:
                return found_sites.pop()
            print("ERROR: Failed to find site, likely ran at the wrong site!")
        if match := re.search(rb"running at site ([^\s]*)\n", self.files.dirac_log):
            found_site = match.group(1)
            print("ERROR: Guessing the site from log:", found_site)
            return found_site.decode()
        else:
            raise NotImplementedError(self, found_sites)

    @cached_property
    def pre_analysis(self) -> dict[str, PreAnalysisResult]:
        """Find input data to the job and determine what action should be taken."""
        # If we don't even have the prodrun config, we can't do anything
        if self.files.steps[0].prodrun_config is None:
            raise NotImplementedError("_process_job_jdl")
        input_lfns = self.files.steps[0].prodrun_config["input"]["files"]
        # We should always have a log file
        if self.files.steps[-1].application_log is None:
            return {lfn: PreAnalysisResult.NO_LOG_FILE for lfn in input_lfns}
        # We should always have a summary XML file
        if self.files.steps[-1].summary_xml is None:
            return {lfn: PreAnalysisResult.NO_XML_SUMMARY for lfn in input_lfns}
        # Check the file statuses from the first job
        files_by_status = {}
        for file in self.files.steps[0].summary_xml.findall(".//input/file"):
            # Entries that don't start with LFN: correspond to file access errors
            # Gaudi can potentially fallback to another replica in this case
            if not file.attrib["name"].startswith("LFN:"):
                continue
            files_by_status.setdefault(file.attrib["status"], []).append(file.attrib["name"])
        if set(files_by_status.keys()) - {"part", "full", "fail"}:
            raise NotImplementedError(self, files_by_status.keys())

        # If the summary XML says the job was successful do some sanity checks
        if ast.literal_eval(self.files.steps[-1].summary_xml.find(".//success").text):
            file_success = all(lfn in files_by_status.get("full", []) for lfn in input_lfns)
            for step in self.files.steps[1:]:
                if step.summary_xml is None:
                    file_success = False
                    break
                if not ast.literal_eval(step.summary_xml.find(".//success").text):
                    file_success = False
                    break
                if len(step.summary_xml.findall(".//input/file")) != 1:
                    raise NotImplementedError("Expected exactly one intermediate input file in the summary XML")
                if step.summary_xml.find(".//input/file").attrib["status"] != "full":
                    file_success = False
                    break

            if file_success:
                if self.files.steps[-1].summary_xml.findall(".//output/file"):
                    # Something probably went wrong in the LHCbDIRAC machinery
                    return {lfn: PreAnalysisResult.UNKNOWN_ERROR_ON_JOB_SUCCESS for lfn in input_lfns}
                if b"LHCbApplicationError" not in self.files.dirac_log:
                    # Ntuples are not considered output files by Gaudi so do some extra sanity checks
                    if b"NTuples saved successfully" in self.files.steps[-1].application_log:
                        return {lfn: PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE for lfn in input_lfns}
                    if self.files.steps[-1].bk_xml is not None:
                        # This can happen if the job fails to contact the TransformationManager
                        # You MUST check the stdout in the output sandbox (i.e. more than logSE)
                        # e.g. 1214907348
                        raise NotImplementedError("Bookkeeping report but no events selected?", self)
                    # TODO: Eventually we should make this impossible by ensuring files always have an FSR
                    # TODO: Once this is done we can make this check depend on the application version
                    return {lfn: PreAnalysisResult.NO_EVENTS_SELECTED for lfn in input_lfns}

        # Try to figure out which input files are bad
        result = {}
        if len(input_lfns) == 1:
            bad_lfns = [input_lfns[0]]
        elif len(self.files.steps) > 1 and len(input_lfns) > 1:
            # If it's the second step that failed we need to reduce the group size to isolate the bad file
            return {lfn: PreAnalysisResult.NEEDS_GROUP_SIZE_REDUCTION for lfn in input_lfns}
        else:
            result |= {lfn: PreAnalysisResult.PROBABLY_OK for lfn in files_by_status.get("full", [])}
            bad_lfns = files_by_status.get("part", [])[:]
            bad_lfns += files_by_status.get("fail", [])
            # If a file is not in the summary XML, we don't know what happened to it
            result |= {
                lfn: PreAnalysisResult.NOT_IN_SUMMARY_XML
                for lfn in input_lfns
                if lfn not in result and lfn not in bad_lfns
            }

        # Bad files need to be analyse more carefully
        return result | {lfn: PreAnalysisResult.NEEDS_LOG_ANALYSIS for lfn in bad_lfns}

    @cached_property
    def log_analysis(self) -> LogFileMatches:
        return AppLogFailureReason.scan_job_log(self)

    @cached_property
    def peak_memory(self) -> int | None:
        if prmon_data := self.files.steps[-1].prmon_data:
            return max(prmon_data["rss+swap"]) * 1024
        return None

    @cached_property
    def problems_by_lfn(self) -> dict[str, LFNDebugInfo]:
        problems_by_lfn = defaultdict(list)
        if len(self.files.steps) == 1 and not set(self.log_analysis.by_lfn).issubset(set(self.pre_analysis)):
            raise NotImplementedError("Found an LFN in log which wasn't expected!?", self)

        if self.files.steps[0].prodrun_config["application"]["name"].startswith("lb-conda"):
            raise LbCondaJobsUnsupported("Debugging lb-conda jobs is not supported", self)

        unique_diagnoses = set(self.pre_analysis.values())

        override_diagnosis = None
        # If all files are successfully processed and we found an error during
        # finalization we need to reduce the group size
        if (
            len(self.log_analysis.by_lfn) > 1
            and unique_diagnoses == {PreAnalysisResult.PROBABLY_OK}
            and self.log_analysis.finalize
        ):
            override_diagnosis = PreAnalysisResult.NEEDS_GROUP_SIZE_REDUCTION

        info_cls = partial(
            LFNDebugInfo,
            log_url=self.url,
            job_id=self.job_id,
            site=self.site,
            peak_memory=self.peak_memory,
            transform_id=self.transform_id,
        )
        found_reason_from_log_analysis = False
        for lfn, diagnosis in reversed(self.pre_analysis.items()):
            diagnosis = override_diagnosis or diagnosis
            if diagnosis == PreAnalysisResult.NOT_IN_SUMMARY_XML and self.log_analysis.by_lfn.get(lfn):
                # The summary XML has lied to us about the file having not been
                # processed, trust the log analysis instead
                diagnosis = PreAnalysisResult.NEEDS_LOG_ANALYSIS
            info_cls = partial(info_cls, lfn=lfn, pre_analysis=diagnosis)
            match diagnosis:
                case PreAnalysisResult.NO_LOG_FILE:
                    problems_by_lfn[lfn] = info_cls()
                case PreAnalysisResult.NO_XML_SUMMARY:
                    if self.log_analysis.init:
                        problems_by_lfn[lfn] = info_cls(maybe_reason=self.log_analysis.init[0])
                    else:
                        problems_by_lfn[lfn] = info_cls()
                case PreAnalysisResult.NEEDS_LOG_ANALYSIS:
                    if self.log_analysis.init:
                        for reason in self.log_analysis.init:
                            if isinstance(reason, WorkerNodeIssue):
                                maybe_reason = [reason]
                                break
                        else:
                            # See if the file is in status part to guess the bad LFN
                            input_files = self.files.steps[-1].summary_xml.findall(".//input/file")
                            if input_files:
                                bad_lfns = [f.attrib["name"] for f in input_files if f.attrib["status"] == "part"]
                            else:
                                bad_lfns = sorted(self.pre_analysis)
                            if len(bad_lfns) != 1:
                                raise NotImplementedError(self)
                            if bad_lfns[0] != lfn:
                                raise NotImplementedError(self)
                            maybe_reason = self.log_analysis.init
                    elif maybe_reason := self.log_analysis.by_lfn.get(lfn):
                        pass
                    elif len(self.files.steps[0].prodrun_config["input"]["files"]) == 1 and (
                        maybe_reason := next(iter(self.log_analysis.by_lfn.values()))
                    ):
                        pass
                    elif self.log_analysis.finalize:
                        if len(self.pre_analysis) == 1:
                            maybe_reason = self.log_analysis.finalize
                        elif diagnosis:
                            pass
                        else:
                            raise NotImplementedError(self)
                    elif not found_reason_from_log_analysis:
                        raise NotImplementedError(f"{self=} {len(self.files.steps[0].application_log)=}")
                    if maybe_reason:
                        found_reason_from_log_analysis = True
                        problems_by_lfn[lfn] = info_cls(maybe_reason=maybe_reason[0])
                    elif not found_reason_from_log_analysis:
                        raise NotImplementedError(self)
                    else:
                        problems_by_lfn[lfn] = info_cls(pre_analysis=PreAnalysisResult.PROBABLY_OK)
                case (
                    PreAnalysisResult.UNKNOWN_ERROR_ON_JOB_SUCCESS
                    | PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE
                    | PreAnalysisResult.NO_EVENTS_SELECTED
                ):
                    if len(unique_diagnoses) != 1:
                        raise NotImplementedError(self)
                    if self.log_analysis.init:
                        raise NotImplementedError(self)
                    if len(self.files.steps) == 1 and set(self.log_analysis.by_lfn) != set(self.pre_analysis):
                        raise NotImplementedError(self)
                    if self.log_analysis.finalize:
                        problems_by_lfn[lfn] = info_cls(maybe_reason=self.log_analysis.finalize[0])
                    else:
                        problems_by_lfn[lfn] = info_cls()

                case PreAnalysisResult.PROBABLY_OK:
                    if maybe_reason := self.log_analysis.by_lfn.get(lfn):
                        if not maybe_reason:
                            raise NotImplementedError(self)
                        problems_by_lfn[lfn] = info_cls(maybe_reason=maybe_reason[0])
                    else:
                        problems_by_lfn[lfn] = info_cls()

                case PreAnalysisResult.NOT_IN_SUMMARY_XML:
                    if self.log_analysis.by_lfn.get(lfn):
                        raise NotImplementedError(self)
                    problems_by_lfn[lfn] = info_cls()

                case PreAnalysisResult.NEEDS_GROUP_SIZE_REDUCTION:
                    if len(unique_diagnoses) != 1:
                        raise NotImplementedError(self)
                    maybe_reason = None
                    if unique_diagnoses == {PreAnalysisResult.PROBABLY_OK} and self.log_analysis.finalize:
                        maybe_reason = self.log_analysis.finalize[0]
                    problems_by_lfn[lfn] = info_cls(maybe_reason=maybe_reason)
                case _:
                    raise NotImplementedError(diagnosis)

        return dict(problems_by_lfn)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import argparse
import asyncio
from pathlib import Path
import re
import shutil


import DIRAC
from DIRAC.Core.Security.Properties import PRODUCTION_MANAGEMENT
from rich.console import Console
from rich.prompt import Prompt
from rich.markdown import Markdown


def main():
    parser = argparse.ArgumentParser(description="Manage LHCbDIRAC production requests")
    subparsers = parser.add_subparsers(dest="command", required=True)

    subparser = subparsers.add_parser("show-requests", help="Show the current status of the production requests")
    subparser.set_defaults(func=show_requests)
    subparser.add_argument("--execute-actions", action="store_true", help="Execute the actions")

    subparser = subparsers.add_parser("update-metadata", help="Update the metadata stored on GitLab")
    subparser.set_defaults(func=update_metadata)

    subparser = subparsers.add_parser("run-checks", help="Run checks on the production requests")
    subparser.set_defaults(func=run_checks)

    subparser = subparsers.add_parser("debug", help="Debug the production requests")
    subparser.set_defaults(
        func=lambda args: asyncio.run(debug(args.prod_id, summarize=not args.no_summarize, output_dir=args.output_dir))
    )
    subparser.add_argument("prod_id", type=int, nargs="+")
    subparser.add_argument("--no-summarize", action="store_true")
    subparser.add_argument("--output-dir", type=Path)

    subparser = subparsers.add_parser("add-test-case", help="Add a test case for the job ID specified")
    subparser.set_defaults(func=add_test_case)
    subparser.add_argument("job_id", type=int)

    args = parser.parse_args()
    args.func(args)


def show_requests(args):
    DIRAC.initialize()
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools import ProdRequestsGitlabRepo
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.monitoring import (
        analyse_active_productions,
        display_table,
        display_actions,
    )

    if args.execute_actions:
        DIRAC.initialize(security_expression=PRODUCTION_MANAGEMENT)
        match Prompt.ask("Update metadata?", choices=["yes", "no"], default="yes"):
            case "yes":
                update_metadata(args)

    repo = ProdRequestsGitlabRepo(with_auth=args.execute_actions)
    last_update, tables_data, actions = analyse_active_productions(repo)
    console = Console()
    console.rule()
    display_table(console, tables_data)
    console.print(f"Last update: {last_update}")
    console.rule()
    display_actions(console, actions, execute=args.execute_actions)


def update_metadata(args):
    DIRAC.initialize(security_expression=PRODUCTION_MANAGEMENT)
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools import ProdRequestsGitlabRepo
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.monitoring import ACTIVE_PRODUCTION_STATES

    repo = ProdRequestsGitlabRepo(with_auth=True)
    repo.poll(do_status_update=True, states=ACTIVE_PRODUCTION_STATES)


def run_checks(args):
    DIRAC.initialize(security_expression=PRODUCTION_MANAGEMENT)
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools import (
        ProdRequestsGitlabRepo,
        OperationsLogbook,
        do_checks,
    )
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.monitoring import (
        analyse_active_productions,
        display_table,
    )

    repo = ProdRequestsGitlabRepo(with_auth=True)
    logbook = OperationsLogbook()
    last_update, tables_data, _ = analyse_active_productions(repo, states=["checking"])

    console = Console()
    console.rule()
    display_table(console, tables_data)
    console.print(f"Last update: {last_update}")
    console.rule()

    request_ids = {
        str(x[1]): x[0]
        for v in tables_data.values()
        for vv in v.values()
        for vvv in vv.values()
        for vvvv in vvv.values()
        for x in vvvv
    }

    asyncio.run(do_checks(console, logbook, repo, request_ids))


async def debug(prod_ids: list[int], *, summarize: bool = True, output_dir: Path | None = None):
    DIRAC.initialize(security_expression=PRODUCTION_MANAGEMENT)

    from rich import print

    from ..Utilities.ProductionTools.LogAnalysis import analyze_prod, summarize_prod, group_by_problem, output
    from ..Utilities.ProductionTools.LogAnalysis.utils import Progress, ProgressUpdater
    from ..Utilities.ProductionTools.LogAnalysis.prod_analyzer import GroupedProblems

    async def show_production(
        transform_id: int, progress_updater, *, summarize: bool = True
    ) -> tuple[int, dict[str, int], GroupedProblems]:
        all_problems = await analyze_prod(transform_id, progress=progress_updater)
        if summarize:
            all_problems = summarize_prod(all_problems)
        return transform_id, group_by_problem(all_problems)

    if output_dir is not None:
        output_dir.mkdir(parents=True, exist_ok=False)

    with Progress() as progress:
        task = progress.add_task("Reading logs", total=None)
        progress_updater = ProgressUpdater(progress, task)
        async with asyncio.TaskGroup() as tg:
            tasks = [
                tg.create_task(show_production(prod_id, progress_updater, summarize=summarize)) for prod_id in prod_ids
            ]
            for coro in asyncio.as_completed(tasks):
                prod_id, grouped = await coro
                raw_markdown = output.to_markdown(grouped, is_summary=summarize)
                print(Markdown(raw_markdown))
                if output_dir:
                    md_path = output_dir / f"{prod_id}.md"
                    md_path.write_text(raw_markdown)
                    print(f"Wrote {md_path}")
                print(f"Analyzed {len(tasks)}")


def add_test_case(args):
    """Add a test case for the job ID specified.

    Downloads job logs, analyzes them, creates fixtures, and prepares a git branch
    for contributing the test case to LHCbDIRAC.

    :param args: Namespace with job_id attribute
    """
    import json
    import subprocess
    import sys
    import tempfile
    from pathlib import Path

    DIRAC.initialize()
    from DIRAC.Interfaces.API.Dirac import Dirac

    console = Console()
    job_id = args.job_id

    # Check if we're in an LHCbDIRAC repository
    try:
        result = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            capture_output=True,
            text=True,
            check=True,
        )
        repo_root = Path(result.stdout.strip())

        # Verify it's actually LHCbDIRAC
        if not (repo_root / "src" / "LHCbDIRAC").exists():
            console.print("[red]Not in LHCbDIRAC repository. Please cd to LHCbDIRAC root.[/red]")
            sys.exit(1)
    except subprocess.CalledProcessError:
        console.print("[red]Not in a git repository. Please cd to LHCbDIRAC root.[/red]")
        sys.exit(1)

    console.print(f"[green]Repository root:[/green] {repo_root}")

    # Get job parameters to find log location
    console.print(f"[blue]Fetching job parameters for {job_id}...[/blue]")
    dirac = Dirac()
    result = dirac.getJobParameters(job_id)

    if not result["OK"]:
        console.print(f"[red]Failed to get job parameters: {result['Message']}[/red]")
        sys.exit(1)

    params = result["Value"]

    # Extract log path from 'Log URL' parameter
    # Format: '<a href="https://lhcb-dirac-logse.web.cern.ch:443/lhcb/MC/2024/LOG/00335034/0000/00000030/">Log file directory</a>'
    # We want: /lhcb/MC/2024/LOG/00335034/0000/00000030/
    log_url_param = params.get("Log URL", "")

    url_match = re.search(r'href="([^"]+)"', log_url_param)
    if not url_match:
        console.print(f"[red]No Log URL found in job parameters[/red]")
        console.print(f"[dim]Available params: {list(params.keys())}[/dim]")
        sys.exit(1)

    log_url = url_match.group(1)
    console.print(f"[green]Log URL:[/green] {log_url}")

    # Extract path from URL (after the hostname)
    # URL format: https://lhcb-dirac-logse.web.cern.ch:443/lhcb/MC/2024/LOG/...
    path_match = re.search(r"https?://[^/]+(/lhcb/.+?)/?$", log_url)
    if not path_match:
        console.print(f"[red]Could not extract path from Log URL: {log_url}[/red]")
        sys.exit(1)

    log_path = path_match.group(1).lstrip("/")

    # Download logs to temporary location
    test_logs_dir = (
        repo_root / "src/LHCbDIRAC/ProductionManagementSystem/Utilities/ProductionTools/LogAnalysis/test/examples/logs"
    )
    job_dir = test_logs_dir / str(job_id)

    if job_dir.exists():
        console.print(f"[yellow]Warning: {job_dir} already exists. Overwriting...[/yellow]")
        shutil.rmtree(job_dir)

    job_dir.mkdir(parents=True, exist_ok=True)

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        zip_file = tmpdir / f"{job_id}.zip"

        # Download with xrdcp
        console.print(f"[blue]Downloading logs from EOS...[/blue]")
        xrd_url = f"root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/logSE/{log_path}.zip"

        try:
            subprocess.run(
                ["xrdcp", xrd_url, str(zip_file)],
                check=True,
                capture_output=True,
            )
        except subprocess.CalledProcessError as e:
            console.print(f"[red]Failed to download logs: {e.stderr.decode()}[/red]")
            sys.exit(1)

        # Unzip
        console.print(f"[blue]Extracting logs...[/blue]")
        subprocess.run(["unzip", "-q", str(zip_file), "-d", str(job_dir)], check=True)

    console.print(f"[green]Logs extracted to {job_dir}[/green]")

    # Move any files inside subfolders up to job_dir
    for subdir in job_dir.iterdir():
        if subdir.is_dir():
            for file in subdir.iterdir():
                file.rename(job_dir / file.name)
            subdir.rmdir()

    # Compress logs to tar.zst
    console.print(f"[blue]Compressing logs to tar.zst...[/blue]")
    tar_zst_file = test_logs_dir / f"{job_id}.tar.zst"

    try:
        # Create tar and pipe to zstd
        with subprocess.Popen(
            ["tar", "-cf", "-", "-C", str(test_logs_dir), str(job_id)],
            stdout=subprocess.PIPE,
        ) as tar_proc:
            with open(tar_zst_file, "wb") as zst_file:
                subprocess.run(
                    ["zstd", "-T0", "-19"],
                    stdin=tar_proc.stdout,
                    stdout=zst_file,
                    check=True,
                )
    except subprocess.CalledProcessError as e:
        console.print(f"[red]Failed to compress: {e}[/red]")
        sys.exit(1)

    console.print(f"[green]Compressed to {tar_zst_file}[/green]")

    # Analyze the job to create the expected JSON output
    console.print(f"[blue]Analyzing job logs...[/blue]")
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.test import generate_fixture_jobid

    # Load job files

    reasons_yaml = (
        repo_root / "src/LHCbDIRAC/ProductionManagementSystem/Utilities/ProductionTools/LogAnalysis/reasons.yaml"
    )

    try:
        generate_fixture_jobid(job_id)
        json_file = test_logs_dir / f"{job_id}.json"
        if not json_file.exists():
            raise RuntimeError("Expected JSON output file not created")
        console.print(f"[green]Analysis output written to {json_file}[/green]")
    except NotImplementedError as e:
        from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.test.Test_log_analysis import (
            UPDATE_REFERENCES_ENV_VAR,
            TEST_MODULE_NAME,
        )

        console.print(f"[red]Log analysis not implemented: {e}[/red]")
        console.print(f"[yellow]Please make the necessary modifications in {reasons_yaml}[/yellow]")
        console.print()
        console.print("[blue]Then run the test with:[/blue]")
        console.print(f"  [yellow]{UPDATE_REFERENCES_ENV_VAR}=1 pytest -k '{TEST_MODULE_NAME} and {job_id}'[/yellow]")
        console.print()
        sys.exit(1)
    except Exception as e:
        console.print(f"[red]Failed to analyze logs: {e}[/red]")
        import traceback

        traceback.print_exc()
        sys.exit(1)
    finally:
        # Always clean up uncompressed directory - tar.zst has all needed data
        if job_dir.exists():
            shutil.rmtree(job_dir)

    # Add files
    console.print(f"[blue]Adding files to git...[/blue]")
    subprocess.run(["git", "add", str(tar_zst_file.relative_to(repo_root))], check=True, cwd=repo_root)
    subprocess.run(["git", "add", str(json_file.relative_to(repo_root))], check=True, cwd=repo_root)

    console.print(f"[green]✓ Test case added successfully![/green]")
    console.print()
    console.print("[blue]Next steps:[/blue]")
    console.print(f"  1. Commit changes: [yellow]git commit -m 'Add log analysis test case for job {job_id}'[/yellow]")
    console.print(f"  3. Push branch: [yellow]git push -u fork <branch>[/yellow]")
    console.print(f"  4. Open a merge request on GitLab")
    console.print()
    console.print(f"[dim]Files: {tar_zst_file.name}, {json_file.name}[/dim]")


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import re
import json
import xml.etree.ElementTree as ET
import fnmatch
from collections import defaultdict
from collections.abc import Sequence
from functools import cached_property
from typing import Any

RE_STEP_ARTIFACT_NAME = re.compile(
    r"(?P<prefix>[A-Za-z0-9_-]+)_(?P<transform_id>[\d]{8})_(?P<task_id>[\d]{8})_(?P<step_idx>[\d]+)\.(?P<suffix>[a-z]+)"
)


class StepArtifacts:
    """A collection of artifacts which all correspond to a single step inside a job"""

    def __init__(self, files: dict[str, bytes]):
        self._files = files

    def get_file(self, pattern: str) -> bytes | None:
        matches = fnmatch.filter(self._files, pattern)
        if len(matches) == 0:
            return None
        elif len(matches) > 1:
            raise NotImplementedError(f"Multiple matches for {pattern}: {matches}")
        return self._files[matches[0]]

    @property
    def application_log(self) -> bytes | None:
        return self.get_file("*.log")

    @property
    def summary_xml(self) -> ET.ElementTree | None:
        if not hasattr(self, "_summary_xml"):
            xml_content = self.get_file("summary*.xml")
            if xml_content:
                self._summary_xml = ET.ElementTree(ET.fromstring(xml_content.decode("utf-8")))
            else:
                self._summary_xml = None
        return self._summary_xml

    @property
    def prodrun_config(self) -> dict[str, Any] | None:
        if not hasattr(self, "_prodrun_config"):
            json_content = self.get_file("prodConf*.json")
            if json_content is None:
                self._prodrun_config = None
            else:
                self._prodrun_config = json.loads(json_content.decode("utf-8"))
        return self._prodrun_config

    @cached_property
    def prmon_data(self) -> dict[str, Sequence[int]] | None:
        raw_data = self.get_file("prmon_*.txt")
        if raw_data is None:
            return None
        headings, *data = raw_data.decode().splitlines()
        headings = headings.split("\t")
        data = dict(zip(headings, zip(*(list(map(int, x.split("\t"))) for x in data))))
        if data:
            data["rss+swap"] = [r + s for r, s in zip(data["rss"], data["swap"])]
        return data

    @property
    def bk_xml(self) -> ET.ElementTree | None:
        if not hasattr(self, "_bk_xml"):
            xml_content = self.get_file("bookkeeping*.xml")
            if xml_content is None:
                self._bk_xml = None
            else:
                self._bk_xml = ET.ElementTree(ET.fromstring(xml_content.decode("utf-8")))
        return self._bk_xml


class JobArtifactConsistencyError(ValueError):
    """Raised when the artifacts of a job are inconsistent"""


class JobArtifacts:
    """A collection of artifacts which all correspond to a single job"""

    @classmethod
    def from_files(cls, files: dict[str, bytes], *, transform_id: int, task_id: int):
        job_files = {}
        steps = defaultdict(dict)
        for filename, contents in files.items():
            if match := RE_STEP_ARTIFACT_NAME.fullmatch(filename):
                steps[match.group("step_idx")][filename] = contents

                # Sanity check that all files in a step have the expected transform and task IDs
                if transform_id != int(match.group("transform_id")):
                    raise JobArtifactConsistencyError("Transformation ID mismatch")
                if task_id != int(match.group("task_id")):
                    raise JobArtifactConsistencyError("Task ID mismatch")
            else:
                job_files[filename] = contents

        steps = [StepArtifacts(files) for _, files in sorted(steps.items(), key=lambda x: x[0])]
        return cls(steps, job_files)

    def __init__(self, steps: list[StepArtifacts], files: dict[str, bytes]):
        self.steps = steps
        self._files = files

    @property
    def dirac_log(self) -> bytes | None:
        return self._files.get("std.out")

    @property
    def workflow(self) -> ET.ElementTree:
        """Parse and return the workflow XML from the job description."""
        if not hasattr(self, "_workflow"):
            job_description = self._files.get("jobDescription.xml")
            if job_description is None:
                self._workflow = None
            else:
                self._workflow = ET.fromstring(job_description.decode("utf-8"))
        return self._workflow

    @property
    def xml_catalog(self) -> ET.ElementTree | None:
        if not hasattr(self, "_xml_catalog"):
            xml_content = self._files.get("pool_xml_catalog.xml")
            if xml_content is None:
                self._xml_catalog = None
            else:
                self._xml_catalog = ET.ElementTree(ET.fromstring(xml_content.decode("utf-8")))
        return self._xml_catalog



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2025 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Check and verify ROOT files from the LHCb grid"""
from __future__ import annotations

import asyncio
from functools import reduce
import importlib.resources
import json
import mmap
import re
import ssl
import sys
import tempfile
import traceback
import zlib
from contextlib import ExitStack, AsyncExitStack
from pathlib import Path

import httpx
import uproot
from rich.progress import (
    Progress,
    BarColumn,
    DownloadColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.ReturnValues import convertToReturnValue

STEP_SIZE = 1000


def parseArgs():
    outputDir = None
    noVerify = False
    fast = False
    excludeSeRegex = None

    @convertToReturnValue
    def setOutputDir(s: str):
        nonlocal outputDir
        outputDir = Path(s)

    @convertToReturnValue
    def setNoVerify(_):
        nonlocal noVerify
        noVerify = True

    @convertToReturnValue
    def setFast(_):
        nonlocal fast
        fast = True

    @convertToReturnValue
    def setExcludeSeRegex(s: str):
        nonlocal excludeSeRegex
        excludeSeRegex = re.compile(s)

    switches = [
        ("", "output-dir=", "Directory to save downloaded files", setOutputDir),
        ("", "no-verify", "Do not verify file contents", setNoVerify),
        ("", "fast", "Use all sources to download the file and don't validate SEs individually", setFast),
        ("", "exclude-se-regex=", "Regex to exclude SEs from download", setExcludeSeRegex),
    ]
    Script.registerSwitches(switches)
    Script.registerArgument("lfn: LFN of the file to check")
    Script.registerArgument("branch: Branch name to check (optional)", mandatory=False)
    Script.registerArgument("event: Event number to check (optional)", mandatory=False)
    Script.parseCommandLine(ignoreErrors=False)
    lfn, branch, event = Script.getPositionalArgs(group=True)

    if event is not None:
        event = int(event)

    return lfn, branch, event, outputDir, noVerify, fast, excludeSeRegex


@Script()
def main():
    lfn, branch, event, outputDir, noVerify, fast, excludeSeRegex = parseArgs()

    with ExitStack() as stack:
        if outputDir:
            destination = outputDir / Path(lfn).name
        else:
            tmp = Path(stack.enter_context(tempfile.TemporaryDirectory()))
            destination = tmp / Path(lfn).name
        asyncio.run(download_file(lfn, destination, validate_all_replicas=not fast, exclude=excludeSeRegex))

        gLogger.notice(f"Downloaded {lfn} to {destination}")
        if noVerify:
            gLogger.notice("Skipping verification")
            return
        f = stack.enter_context(uproot.open(destination))

        err_msg = None
        bad_streamers = len(f._file.streamers) == 0
        if bad_streamers:
            # example: /lhcb/LHCb/Collision24/RDLOW.DST/00276117/0003/00276117_00036205_1.rdlow.dst
            err_msg = "No streamers found"
            gLogger.warn(f"{lfn} has no streamers, probably corrupted")

        tree_names = f.keys(recursive=True, cycle=False, filter_classname="TTree")
        if "Event" in tree_names and branch:
            tree_names = ["Event"]
        try:
            for tree_name in tree_names:
                gLogger.info(f"{tree_name=}")
                check_file(f, tree_name, branch, event)
        except Exception as e:
            traceback.print_exc()
            err_msg = repr(e)
            if branch is not None:
                gLogger.error(f"Affected branch {branch}")
        else:
            if not bad_streamers:
                gLogger.notice(f"{lfn} appears to be valid")
                return

        gLogger.error(f"Error reading {lfn} {err_msg}")

        module_name = "LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis"
        with importlib.resources.files(module_name).joinpath("known_corruption.json").open("r") as known_file:
            metadata = json.load(known_file)

        gLogger.error(f"Found corruption in {lfn} {branch} {event}:\n{err_msg}")
        known_message = metadata.get(lfn, {}).get(branch, {}).get(str(event))
        if known_message:
            if err_msg == known_message:
                gLogger.notice("Message matches known corruption")
            else:
                gLogger.warn(f"Known corruption with different message: {known_message}")
        else:
            gLogger.error(f"Unknown corruption: {json.dumps(err_msg)}")
        sys.exit(1)


class Adler32:
    """Adler32 checksum calculator."""

    def __init__(self):
        self._checksum = 1
        # Need to track length for combining checksums
        self._length = 0

    def __repr__(self):
        return f"Adler32(checksum={self.checksum:#010x}, length={self.length})"

    def update(self, data):
        self._checksum = zlib.adler32(data, self._checksum)
        self._length += len(data)

    @property
    def checksum(self):
        return self._checksum & 0xFFFFFFFF

    @property
    def length(self):
        return self._length

    def __str__(self):
        return f"{self.checksum:#010x}"

    def __and__(self, other: Adler32) -> Adler32:
        if not isinstance(other, Adler32):
            return NotImplemented
        combined = Adler32()
        combined._checksum = self._combine_adler32(other.checksum, other.length)
        combined._length = self.length + other.length
        return combined

    def _combine_adler32(self, adler2: int, len2: int) -> int:
        """
        Combine two Adler-32 checksums.
        adler1: checksum of first block
        adler2: checksum of second block
        len2: length of second block
        """
        MOD = 65521

        # Extract A and B from checksums
        A1 = self.checksum & 0xFFFF
        B1 = (self.checksum >> 16) & 0xFFFF
        A2 = adler2 & 0xFFFF
        B2 = (adler2 >> 16) & 0xFFFF

        # Combine
        A = (A1 + A2 - 1) % MOD
        B = (B1 + B2 + len2 * A1 - len2) % MOD

        return (B << 16) | A


class MmapFile:
    """Context manager for memory-mapped file access."""

    def __init__(self, path: Path, size: int):
        self.path = path
        self.size = size
        self._fh = None
        self._mm = None

    def __enter__(self):
        self._fh = open(self.path, "w+b")
        self._fh.truncate(self.size)
        self._mm = mmap.mmap(self._fh.fileno(), 0)
        return self._mm

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._mm is not None:
            self._mm.flush()
            self._mm.close()
        if self._fh is not None:
            self._fh.close()
        return False


class DownloadWorkerCallback:
    def __init__(self, mm, start, progress_callback, cancel_callback=None, completion_callback=None):
        self.mm = mm
        self.current_pos = start
        self.progress_callback = progress_callback
        self.cancel_callback = cancel_callback
        self.completion_callback = completion_callback
        self.bytes_downloaded = 0

    def handle_chunk(self, chunk: bytes):
        self.mm[self.current_pos : self.current_pos + len(chunk)] = chunk
        self.current_pos += len(chunk)
        self.bytes_downloaded += len(chunk)
        self.progress_callback(len(chunk))

    def handle_cancel(self):
        if self.cancel_callback:
            self.cancel_callback(self.bytes_downloaded)

    def handle_completion(self):
        if self.completion_callback:
            self.completion_callback()


class ChunkedDownloaderWorker:
    def __init__(self, client: httpx.AsyncClient, se: str, pfn: str, queue: asyncio.Queue, task_id: int):
        self.client = client
        self.se = se
        self.pfn = pfn
        self.queue = queue
        self.task_id = task_id
        self.queue.put_nowait(self)

    def __repr__(self) -> str:
        return f"ChunkedDownloaderWorker(se={self.se}, pfn={self.pfn})"

    async def download(self, callback: DownloadWorkerCallback, *, byte_range: tuple[int, int] | None) -> Adler32:
        checksum = Adler32()
        headers = {}
        expected_length = None
        if byte_range is not None:
            self.current_start, self.current_end = byte_range
            headers["Range"] = f"bytes={self.current_start}-{self.current_end-1}"
            gLogger.debug(f"{self.se} downloading range {self.current_start}-{self.current_end-1}")
            expected_length = self.current_end - self.current_start
        try:
            async with AsyncExitStack() as stack:
                coro = self.client.stream("GET", self.pfn, headers=headers)
                response = await stack.enter_async_context(coro)
                content_length = int(response.headers.get("Content-Length", 0))
                if expected_length is not None and content_length != expected_length:
                    msg = f"Unexpected content length for range from {self!r}"
                    raise NotImplementedError(msg)
                response.raise_for_status()
                async for chunk in response.aiter_bytes(chunk_size=32_768):
                    checksum.update(chunk)
                    callback.handle_chunk(chunk)
        except asyncio.CancelledError:
            gLogger.debug(f"{self.se} download was cancelled")
            callback.handle_cancel()
            raise
        else:
            callback.handle_completion()
        finally:
            self.queue.put_nowait(self)
        return checksum


async def download_file(lfn, dest, *, validate_all_replicas: bool = False, exclude: re.Pattern | None = None):
    from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
    from DIRAC.Core.Security.Locations import getCAsLocation, getProxyLocation
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

    replicas = returnValueOrRaise(DataManager().getReplicas(lfn, diskOnly=True, protocol="https"))
    if replicas["Failed"]:
        raise NotImplementedError(f"Failed to get replicas for {lfn}: {replicas['Failed']}")

    result = returnValueOrRaise(FileCatalog().getFileMetadata(lfn))
    if result["Failed"]:
        raise NotImplementedError(f"Failed to get metadata for {lfn}: {result['Failed']}")
    expected_checksum = int(result["Successful"][lfn]["Checksum"], base=16)
    expected_size = result["Successful"][lfn]["Size"]

    ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, capath=getCAsLocation())
    ctx.load_cert_chain(getProxyLocation())

    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(httpx.AsyncClient(timeout=None, verify=ctx, follow_redirects=True))

        progress_columns = [
            TextColumn("[bold blue]{task.fields[url]}", justify="right"),
            BarColumn(),
            TransferSpeedColumn(),
        ]
        progress_columns.extend([DownloadColumn(), TimeElapsedColumn(), TimeRemainingColumn()])
        progress = Progress(*progress_columns, transient=False)

        progress = stack.enter_context(progress)

        if validate_all_replicas:
            tasks = []
            for se, pfn in replicas["Successful"][lfn].items():
                if exclude and exclude.fullmatch(se):
                    gLogger.info(f"Excluding {se} due to exclude regex")
                    continue
                task_id = progress.add_task("download", url=se, total=0)
                tasks.append(download_replica(client, se, pfn, dest, progress, task_id))
                # Later replicas should only compute the checksum
                dest = None
            checksums = await asyncio.gather(*tasks)
        else:
            gLogger.notice(f"Downloading file using all available replicas to {dest}")
            mm = stack.enter_context(MmapFile(dest, expected_size))

            def progress_callback(progress: Progress, summary_task_id: int, task_id: int):
                def callback(n: int):
                    progress.update(summary_task_id, advance=n)
                    progress.update(task_id, advance=n)

                return callback

            def cancel_callback(progress: Progress, summary_task_id: int, task_id: int):
                def callback(bytes_downloaded: int):
                    progress.update(summary_task_id, advance=-bytes_downloaded)
                    progress.update(task_id, advance=-bytes_downloaded)

                return callback

            current_pos = 0
            chunk_size = 16 * 1024 * 1024

            workers = []
            available_workers = asyncio.Queue()
            summary_task_id = progress.add_task("summary", url="Overall", total=expected_size)

            for se, pfn in replicas["Successful"][lfn].items():
                if exclude and exclude.fullmatch(se):
                    gLogger.info(f"Excluding {se} due to exclude regex")
                    continue
                task_id = progress.add_task("download", url=se, total=0)
                worker = ChunkedDownloaderWorker(client, se, pfn, available_workers, task_id)
                workers.append(worker)

            tasks = {}
            async with asyncio.TaskGroup() as tg:
                while current_pos < expected_size:
                    worker = await available_workers.get()
                    start = current_pos
                    end = min(current_pos + chunk_size, expected_size)

                    pcallback = progress_callback(progress, summary_task_id, worker.task_id)
                    ccallback = cancel_callback(progress, summary_task_id, worker.task_id)
                    byte_range = (start, end)
                    callback = DownloadWorkerCallback(mm, byte_range[0], pcallback, cancel_callback=ccallback)
                    task = tg.create_task(worker.download(callback, byte_range=byte_range))
                    tasks[byte_range] = [task]
                    current_pos = end

                # Add alternative sources for the data which is still being downloaded to prevent slow sources
                # from blocking the download
                while True:
                    worker = await available_workers.get()
                    gLogger.debug(f"Got available worker {worker.se}")
                    gLogger.debug(f"Currently have {len(tasks)} download tasks")
                    active_ranges = [k for k, ts in tasks.items() if not any(t.done() for t in ts)]
                    gLogger.debug(f"Currently have {len(active_ranges)} active download tasks")
                    for k, ts in tasks.items():
                        if k not in active_ranges:
                            continue
                        for t in ts:
                            if not t.done():
                                gLogger.debug(f"  {k} ({len(ts)=}): Task {t.get_name()} is active")
                            elif t.done():
                                gLogger.debug(f"  {k} ({len(ts)=}): Task {t.get_name()} is done")
                            elif t.cancelling():
                                gLogger.debug(f"  {k} ({len(ts)=}): Task {t.get_name()} is cancelling")
                            else:
                                raise NotImplementedError("Unknown task state")
                    if not active_ranges:
                        break
                    # As chunks are downloaded sequentially, the task with the slowest chunk will be the one
                    # with the lowest start position
                    byte_range = min(active_ranges)
                    gLogger.debug(f"Adding additional source ({worker.se}) for range {byte_range=}")
                    pcallback = progress_callback(progress, summary_task_id, worker.task_id)
                    ccallback = cancel_callback(progress, summary_task_id, worker.task_id)

                    async def completion_callback(tasks=tasks[byte_range]):
                        for t in tasks:
                            if not t.done():
                                gLogger.debug(f"Cancelling duplicate download task {t.get_name()}")
                                t.cancel()

                    callback = DownloadWorkerCallback(
                        mm,
                        byte_range[0],
                        pcallback,
                        cancel_callback=ccallback,
                        completion_callback=lambda: tg.create_task(completion_callback()),
                    )
                    task = tg.create_task(worker.download(callback, byte_range=byte_range))
                    tasks[byte_range].append(task)

            # Keep only the finished task for each byte range
            for byte_range, ts in tasks.items():
                tasks[byte_range] = [t for t in ts if not t.cancelled()][0]

            # Combine the checksums from all chunks
            full_checksum = reduce(lambda a, b: a & b, (t.result() for _, t in sorted(tasks.items())))
            checksums = [("Multi-source", full_checksum)]

    if len(checksums) == 0:
        raise NotImplementedError(f"No replicas found for {lfn}")

    for se, checksum in checksums:
        gLogger.info(f"Checksum for {lfn} at {se}: {checksum} == {expected_checksum:#010x}")
        if checksum.checksum != expected_checksum:
            raise NotImplementedError(f"Checksum mismatch for {se}: {checksum} != {expected_checksum:#010x}")
    gLogger.notice("All checksums match!")


async def download_replica(
    client: httpx.AsyncClient, se: str, url: str, dest_path: str | None, progress: Progress, task_id: int
) -> Adler32:
    """Download a single URL to disk, updating the given Rich task."""
    checksum = Adler32()

    async with AsyncExitStack() as stack:
        response = await stack.enter_async_context(client.stream("GET", url))
        response.raise_for_status()
        total = int(response.headers.get("Content-Length", 0))
        progress.update(task_id, total=total)

        out_file = stack.enter_context(open(dest_path, "wb")) if dest_path else None

        async for chunk in response.aiter_bytes(chunk_size=32_768):
            checksum.update(chunk)
            if out_file:
                out_file.write(chunk)
            progress.update(task_id, advance=len(chunk))

    return se, checksum


def check_file(f, tree_name, branch, event):
    interpretation = None
    if tree_name in ["Event", "FileRecords"]:
        interpretation = uproot.interpretation.jagged.AsJagged(uproot.interpretation.numerical.AsDtype("u1"))
    start = None
    if branch and event:
        f[tree_name][branch].array(interpretation, entry_start=event, entry_stop=event + 1, library="np")
    elif branch:
        for start in range(0, f[tree_name].num_entries, STEP_SIZE):
            f[tree_name][branch].array(interpretation, entry_start=start, entry_stop=start + STEP_SIZE, library="np")
    else:
        for branch in f[tree_name].keys():
            gLogger.info(f"Checking {tree_name}/{branch}")
            for start in range(0, f[tree_name].num_entries, STEP_SIZE):
                f[tree_name][branch].array(
                    interpretation, entry_start=start, entry_stop=start + STEP_SIZE, library="np"
                )


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Provides access to Analysis Productions specific functionality

For more information see :py:mod:`.AnalysisProductionsClient`.
"""
from datetime import datetime
from itertools import chain

from DIRAC import gConfig, gLogger
from DIRAC.Core.Tornado.Server.TornadoService import TornadoService
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise, convertToReturnValue
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Storage.StorageElement import StorageElement
from LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB import OracleBookkeepingDB
from LHCbDIRAC.TransformationSystem.DB.TransformationDB import TransformationDB
from LHCbDIRAC.ProductionManagementSystem.DB.AnalysisProductionsDB import AnalysisProductionsDB
from LHCbDIRAC.ProductionManagementSystem.DB.AnalysisProductionsObjects import ActionsInputModel


optDate = (type(None), datetime)
optString = (type(None), str)
optList = (type(None), list)
optBool = (type(None), bool)

sLog = gLogger.getSubLogger(__name__)
sLog._setOption("tornadoComponent", "ProductionManagement/TornadoAnalysisProduction")


class TornadoAnalysisProductionsHandler(TornadoService):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """initialize handler"""
        cls._db = AnalysisProductionsDB(parentLogger=cls.log)
        cls._transDB = TransformationDB(parentLogger=cls.log)
        cls._bkDB = OracleBookkeepingDB(parentLogger=cls.log)

    types_listAnalyses = [optDate]

    @convertToReturnValue
    def export_listAnalyses(self, at_time):
        """See :meth:`~.AnalysisProductionsClient.listAnalyses`"""
        return self._db.listAnalyses(at_time=at_time)

    types_listAnalyses2 = [optDate]

    @convertToReturnValue
    def export_listAnalyses2(self, at_time):
        """See :meth:`~.AnalysisProductionsClient.listAnalyses2`"""
        return self._db.listAnalyses2(at_time=at_time)

    types_listRequests = [optDate]

    @convertToReturnValue
    def export_listRequests(self):
        """See :meth:`~.AnalysisProductionsClient.listRequests`"""
        return self._db.listRequests()

    types_getUsers = [optString]

    @convertToReturnValue
    def export_getUsers(self, username):
        """See :meth:`~.AnalysisProductionsClient.getUsers`"""
        return self._db.getUsers(username=username)

    types_updateUsers = [list]

    @convertToReturnValue
    def export_updateUsers(self, users):
        """See :meth:`~.AnalysisProductionsClient.updateUsers`"""
        return self._db.updateUsers(users=users)

    types_getOwners = [str, str, optDate]

    @convertToReturnValue
    def export_getOwners(self, wg, analysis, at_time):
        """See :meth:`~.AnalysisProductionsClient.getOwners`"""
        return self._db.getOwners(wg=wg, analysis=analysis, at_time=at_time)

    types_setOwners = [str, str, list]

    @convertToReturnValue
    def export_setOwners(self, wg, analysis, owners):
        """See :meth:`~.AnalysisProductionsClient.setOwners`"""
        return self._db.setOwners(wg=wg, analysis=analysis, owners=owners)

    types_getOwnershipHistory = [str, str]

    @convertToReturnValue
    def export_getOwnershipHistory(self, wg: str, analysis: str):
        """See :meth:`~.AnalysisProductionsClient.getOwnershipHistory`"""
        return self._db.getOwnershipHistory(wg=wg, analysis=analysis)

    types_getProductions = [
        optString,
        optString,
        optString,
        optString,
        optString,
        bool,
        bool,
        bool,
        optDate,
        optBool,
        optBool,
    ]

    @convertToReturnValue
    def export_getProductions(
        self,
        wg,
        analysis,
        version,
        name,
        state,
        with_lfns,
        with_pfns,
        with_transformations,
        at_time,
        show_archived=False,
        require_has_publication=False,
    ):
        """See :meth:`~.AnalysisProductionsClient.getProductions`"""
        if (analysis or name or version) and wg is None:
            raise TypeError("wg must be specified when providing an analysis name or sample name/version")
        if with_pfns and not with_lfns:
            raise TypeError("with_pfns cannot be used without with_lfns")
        results = self._db.getProductions(
            wg=wg,
            analysis=analysis,
            version=version,
            name=name,
            state=state,
            at_time=at_time,
            show_archived=show_archived,
            require_has_publication=require_has_publication,
        )
        return _queryToResults(
            results, with_lfns, with_pfns, with_transformations, bkdb=self._bkDB, trfdb=self._transDB
        )

    types_getArchivedRequests = [optString, bool, bool, bool]

    @convertToReturnValue
    def export_getArchivedRequests(self, state, with_lfns, with_pfns, with_transformations):
        """See :meth:`~.AnalysisProductionsClient.getArchivedRequests`"""
        if with_pfns and not with_lfns:
            raise TypeError("with_pfns cannot be used without with_lfns")
        results = self._db.getArchivedRequests(state=state)
        return _queryToResults(
            results, with_lfns, with_pfns, with_transformations, bkdb=self._bkDB, trfdb=self._transDB
        )

    types_getTags = [optString, optString, optDate]

    @convertToReturnValue
    def export_getTags(self, wg, analysis, at_time):
        """See :meth:`~.AnalysisProductionsClient.getTags`"""
        return self._db.getTags(wg, analysis, at_time=at_time)

    types_getKnownAutoTags = []

    @convertToReturnValue
    def export_getKnownAutoTags(self):
        """See :meth:`~.AnalysisProductionsClient.getKnownAutoTags`"""
        return list(self._db.getKnownAutoTags())

    types_registerTransformations = [dict]

    @convertToReturnValue
    def export_registerTransformations(self, transforms):
        """See :meth:`~.AnalysisProductionsClient.registerTransformations`"""
        transforms = {int(k): v for k, v in transforms.items()}
        return self._db.registerTransformations(transforms)

    types_deregisterTransformations = [dict]

    @convertToReturnValue
    def export_deregisterTransformations(self, tIDs):
        """See :meth:`~.AnalysisProductionsClient.deregisterTransformations`"""
        tIDs = {int(k): v for k, v in tIDs.items()}
        return self._db.deregisterTransformations(tIDs)

    types_registerRequests = [list]

    @convertToReturnValue
    def export_registerRequests(self, requests):
        """See :meth:`~.AnalysisProductionsClient.registerRequests`"""
        results = self._db.registerRequests(requests)
        return _queryToResults(
            results, with_lfns=False, with_pfns=False, with_transformations=False, bkdb=self._bkDB, trfdb=self._transDB
        )

    types_addRequestsToAnalysis = [str, str, list[tuple[int, str]]]

    @convertToReturnValue
    def export_addRequestsToAnalysis(self, wg, analysis, requests):
        """See :meth:`~.AnalysisProductionsClient.addRequestsToAnalysis`"""
        return self._db.addRequestsToAnalysis(wg, analysis, requests)

    types_archiveSamples = [list]

    @convertToReturnValue
    def export_archiveSamples(self, sample_ids):
        """See :meth:`~.AnalysisProductionsClient.archiveSamples`"""
        return self._db.archiveSamples(sample_ids)

    types_archiveSamplesAtSpecificTime = [list, datetime]

    @convertToReturnValue
    def export_archiveSamplesAtSpecificTime(self, sample_ids, archive_time):
        """See :meth:`~.AnalysisProductionsClient.archiveSamplesAtSpecificTime`"""
        return self._db.archiveSamplesAtSpecificTime(sample_ids, archive_time)

    types_setState = [dict]

    @convertToReturnValue
    def export_setState(self, newState):
        """See :meth:`~.AnalysisProductionsClient.setState`"""
        newState = {(int(k), ft): vv for k, v in newState.items() for ft, vv in v.items()}
        return self._db.setState(newState)

    types_setTags = [dict, dict]

    @convertToReturnValue
    def export_setTags(self, oldTags, newTags):
        """See :meth:`~.AnalysisProductionsClient.setTags`"""
        return self._db.setTags(oldTags, newTags)

    types_delayHousekeepingInteractionDue = [list, datetime]

    @convertToReturnValue
    def export_delayHousekeepingInteractionDue(self, samples, next_interaction_due):
        """See :meth:`~.AnalysisProductionsClient.delayHousekeepingInteractionDue`"""
        return self._db.delayHousekeepingInteractionDue(samples, next_interaction_due)

    types_getHousekeepingInteractionDueNow = []

    @convertToReturnValue
    def export_getHousekeepingInteractionDueNow(self):
        """See :meth:`~.AnalysisProductionsClient.getHousekeepingInteractionDueNow`"""
        return self._db.getHousekeepingInteractionDueNow()

    types_addPublication = [list, str]

    @convertToReturnValue
    def export_addPublication(self, samples, number):
        """See :meth:`~.AnalysisProductionsClient.addPublication`"""
        return self._db.addPublication(samples, number)

    types_getPublications = [optList]

    @convertToReturnValue
    def export_getPublications(self, sample_ids):
        """See :meth:`~.AnalysisProductionsClient.getPublications`"""
        return self._db.getPublications(sample_ids)

    types_getOpsOverview = []

    @convertToReturnValue
    def export_getOpsOverview(self, category="AP"):
        """See :meth:`~.AnalysisProductionsClient.getOpsOverview`"""
        return returnValueOrRaise(self._transDB.getOpsOverview(category))

    types_getIdsMapping = []

    @convertToReturnValue
    def export_getIdsMapping(self):
        """See :meth:`~.AnalysisProductionsClient.getIdsMapping`"""
        return self._db.getIdsMapping()

    @convertToReturnValue
    def export_commitBulkDatasetActions(self, wg: str, analysis: str, body: dict):
        """See :meth:`~.AnalysisProductionsClient.commitBulkDatasetActions`"""
        body_validated = ActionsInputModel.model_validate(body)
        return self._db.commitBulkDatasetActions(wg, analysis, body_validated)


def _queryToResults(
    results, with_lfns, with_pfns, with_transformations, *, bkdb: OracleBookkeepingDB, trfdb: TransformationDB
):
    """Convert a query of Request or AP objects to a JSON compatible object"""
    p2tID = {
        (result["request_id"], result["filetype"]): {tInfo["id"]: tInfo["used"] for tInfo in result["transformations"]}
        for result in results
    }

    if with_lfns:
        lfns = _getOutputLFNs(p2tID, bkdb=bkdb)
        if with_pfns:
            replicas = _getReplicas(list(chain(*lfns.values())))

        for result in results:
            result["lfns"] = {} if with_pfns else []
            result["total_bytes"] = 0
            if with_pfns:
                result["available_bytes"] = 0
            for tInfo in result["transformations"]:
                for lfn, lfnMeta in lfns.get(tInfo["id"], {}).items():
                    result["total_bytes"] += lfnMeta["FileSize"]
                    if not with_pfns:
                        result["lfns"].append(lfn)
                    # TODO: Remove when https://github.com/pylint-dev/pylint/issues/9628 is fixed
                    elif lfn in replicas["Failed"]:  # pylint: disable=possibly-used-before-assignment
                        result["lfns"][lfn] = []
                        sLog.warn(
                            "Failed to get replicas for Analysis Production LFN",
                            "%s in rID=%s tID=%s: %s"
                            % (lfn, result["request_id"], tInfo["id"], replicas["Failed"][lfn]),
                        )
                    else:
                        result["lfns"][lfn] = [
                            pfn
                            for pfn in replicas["Successful"][lfn].values()
                            # Ignore entries with the value True as these are at unsupported SEs
                            if isinstance(pfn, str)
                        ]
                        if result["lfns"][lfn]:
                            result["available_bytes"] += lfnMeta["FileSize"]

    if with_transformations:
        tIDs = list(set(chain(*p2tID.values())))
        if tIDs:
            retVal = returnValueOrRaise(
                trfdb.getTransformations(
                    condDict={"TransformationID": tIDs},
                    limit=None,
                    columns=["TransformationID", "Status", "Type"],
                )
            )
            extraTransInfos = {t["TransformationID"]: t for t in retVal}

            missing = set(tIDs) - set(extraTransInfos)
            if missing:
                sLog.warn(
                    "_queryToResults (TornadoAnalysisProductionsHandler): Transformations missing from TransformationDB",
                    f"{sorted(missing)}",
                )

            # Only fetch input_queries for transformations that don't have them cached
            tIDs_needing_input_query = [
                tInfo["id"]
                for r in results
                for tInfo in r["transformations"]
                if not tInfo.get("input_query")
                and extraTransInfos.get(tInfo["id"], {}).get("Type") not in ["DerriviedAnalysisProduction"]
            ]

            if tIDs_needing_input_query:
                inputQueries = returnValueOrRaise(trfdb.getBookkeepingQueries(transIDs=tIDs_needing_input_query))
            else:
                inputQueries = {}

            for tInfo in chain(*(r["transformations"] for r in results)):
                extra = extraTransInfos.get(tInfo["id"], None)
                if not extra:
                    tInfo["status"] = "Unknown"
                    tInfo.setdefault("input_query", {})
                    continue

                tInfo["status"] = extra["Status"]
                # Only use fetched input_query if not already cached
                if not tInfo.get("input_query"):
                    tInfo["input_query"] = inputQueries.get(tInfo["id"], {})
    else:
        for result in results:
            del result["transformations"]

    return results


def _getOutputLFNs(pID2tID, *, bkdb):
    sLog.info("Getting output LFNs")
    keys = [(tID, ft) for (_, ft), tIDs in pID2tID.items() for tID, used in tIDs.items() if used]
    lfns = {}
    if not keys:
        return lfns
    retVal = bkdb.getProductionFilesBulk(
        *map(list, zip(*keys, strict=True)),
        "ALL",
    )
    for tID, lfnMetadata in returnValueOrRaise(retVal).items():
        lfns[int(tID)] = {
            lfn: meta
            for lfn, meta in lfnMetadata.items()
            if meta["GotReplica"] and meta["GotReplica"].lower().startswith("y")
        }
    return lfns


def _getReplicas(lfns):
    if not lfns:
        return {"Successful": {}, "Failed": {}}
    sLog.info("Getting replicas for", f"{len(lfns)} LFNs")
    replicas = returnValueOrRaise(DataManager().getReplicas(lfns, diskOnly=True, getUrl=False))
    for se in gConfig.getValue("/Resources/Sites/LCG/LCG.CERN.cern/SE", []):
        seLFNs = {lfn for lfn, replicaInfo in replicas["Successful"].items() if se in replicaInfo}
        if not seLFNs:
            continue
        sLog.info("Getting PFNs for", f"{len(seLFNs)} LFNs at {se}")
        result = returnValueOrRaise(StorageElement(se).getURL(list(seLFNs), protocol="root"))
        if result["Failed"]:
            raise NotImplementedError(result["Failed"])
        for lfn, pfn in result["Successful"].items():
            replicas["Successful"][lfn][se] = pfn
    return replicas



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File

from DIRAC.TransformationSystem.Client.BodyPlugin.BaseBody import BaseBody

BUFFER_TYPE = "-BUFFER"
DST_TYPE = "-DST"


ALREADY_REPLICATED_TARGET = "ALREADY_REPLICATED"


class MoveAPToArchiveBody(BaseBody):
    """
    This body is to be used for the MoveAPToArchive plugin:

    The file either needs to be replicated to an archive and then removed from CERN-ANAPROD,
    or it has already been replicated and only the removal is needed.
    In the later case, the TargetSE will be

    This workflow now also works with DST instead of buffer
    """

    # This is needed to know how to serialize the object
    _attrToSerialize = []

    @staticmethod
    def _createOperation(operationType, lfns, targetSE, sourceSE=None):
        """
        Generate an Operation object of the given Type, with the
        specified targetSE and sourceSE. Associate to it File objects
        from the lfns list

        :param str operationType: Type of the operation (RemoveReplica, ReplicateAndRegister, etc)
        :param list lfns: list of LFNs on which to perform the operation
        :param str targetSE: SE name(s) targeted by the ops
        :param str sourceSE: SE name of the source
        """
        newOp = Operation()
        newOp.Type = operationType
        newOp.TargetSE = targetSE
        if sourceSE:
            newOp.SourceSE = sourceSE
        for lfn in lfns:
            opFile = File()
            opFile.LFN = lfn
            newOp.addFile(opFile)
        return newOp

    def taskToRequest(self, taskID, task, transID):
        """
        Create the request object from the task Dict


        """

        if isinstance(task["InputData"], list):
            lfns = task["InputData"]
        elif isinstance(task["InputData"], str):
            lfns = task["InputData"].split(";")
        else:
            raise ValueError(f"InputData must be a list or a string, not {type(task['InputData'])}")

        req = Request()

        targetSEs = task["TargetSE"].split(",")
        assert len(targetSEs) == 1, targetSEs

        # If not already replicated, replicate to the Archive
        if targetSEs[0] != ALREADY_REPLICATED_TARGET:
            repArchive = self._createOperation("ReplicateAndRegister", lfns, targetSEs)
            req.addOperation(repArchive)

        # Finally, remove the replicas from CERN-DAQ-EXPORT
        anaProdRemoval = self._createOperation("RemoveReplica", lfns, "CERN-ANAPROD")
        req.addOperation(anaProdRemoval)

        return req
